header {*A new primitive*}

theory Lsdest = HOL4ExtraProb:

ML {* ProofContext.thms_containing_limit := 1000;*}

text{* It is time to introduce the example application that is being
  formalized in section \ref{sec:fmm}. We will be looking at the most
  simplistic possible program for finding a satisfying assignment for
  a propositional formula in conjunctive normal form where any clause
  consists of exactly $k$ literals. This problem is known as $k$-SAT. Our
  algorithm simply selects a random assignment for all of the $n$
  variables. We are interested in the probability that the assignment
  fails to satisfy a given clause. The reasons behind this will become
  clear in a while. 

  In the previous section it was stated that one should be able to
  construct any randomized functional program from the three primitive
  building blocks defined there. Of course, this also holds for the
  program we have in mind. Nevertheless, when you try following the
  style these constructs suggest, taking one random bit at a time and
  evaluating somewhere in between, you run into problems. That is to
  say, the clauses are not independent in general. A variable may
  appear in several clauses, and it would be wrong to fetch a new bit
  from the stream every time it is evaluated. Ergo, the simplest way
  to perform the evaluation of a clause independently from the rest is to
  get a list of all $n$ random bits beforehand.

  A function accomplishing this is not hard to devise. It is
  elementary enough to possibly support a lot of programs.*}

types 'a seq = "nat \<Rightarrow> 'a"

consts 
  lsdest:: "nat \<Rightarrow> 'a seq \<Rightarrow> ('a list * 'a seq)"

primrec 
  "lsdest 0 = UNIT []"
  "lsdest (Suc n) = BIND sdest (\<lambda>x. BIND (lsdest n) (\<lambda>l. UNIT (x#l)))"

constdefs
  P:: "((nat \<Rightarrow> bool) \<Rightarrow> bool) \<Rightarrow> real"
  "P \<equiv> prob bern"

declare stake_def[simp] sdrop_def[simp] I_THM[simp] P_def[simp]


lemma lsdest_length: "\<And>s. length (fst (lsdest n s)) = n"
  by (induct n, simp_all add: BIND_def UNIT_def) (unfold split_def, simp)+

lemma lsdest_indep: "IN (lsdest n) indep_fn"
  by (induct n) (auto simp add: INDEP_FN_UNIT INDEP_FN_SDEST INDEP_FN_BIND)

lemma lsdesttakedrop: "\<And>s. lsdest n s = (stake n s, sdrop n s)"
  by (induct n) (simp_all add: UNIT_def BIND_def sdest_def)

text{*We need to know more than these facts to make use of the new
  primitive. The real challenge here
  lies in getting to the probability distribution. Mind that it is not
  sufficient to know the probability of a whole list or a single bit. To
  draw conclusions about clauses, we need assertions about any $k$
  positions in a list of length $n$. On the other hand, this furnishes
  the whole common probability distribution. A few lemmata pave the way. *}

lemma sdrop_stl: "\<And>s. sdrop n (stl s) = stl (sdrop n s)"
(*<*)proof (induct n)
  case 0 show ?case by simp
next
  case (Suc n s)
  hence "sdrop n (stl (s \<circ> Suc)) = stl (sdrop n (s \<circ> Suc))" .
  thus ?case by simp
qed(*>*)


lemma stake_shd:
  "\<And>s. stake u s @ [shd (sdrop u s)] = shd s # stake u (stl s)"
  by (induct u) simp_all

lemma GSPEC_lambda: "GSPEC (\<lambda>s. (s, f s)) = (\<lambda>s. f s)"
  by (simp add: GSPEC_DEF_ALT)

lemma prob_bern_single: "P (\<lambda>s. shd (sdrop n s)) = 1/2"
  using PROB_BERN_SINGLE by (simp add: GSPEC_lambda)
  
lemma shd_sdrop_prob: "P (\<lambda>s. shd (sdrop n s)=b) = 1/2" 
proof (cases b)
  case True
  thus ?thesis using prob_bern_single
    by simp
next
  case False
  have "IN (Not \<circ> shd) (events bern)"
    by (simp only: EVENTS_BERN_SHD)
  hence "IN (\<lambda>s.  shd s = False) (events bern)"
    by (simp add: comp_def)
  also have "(IN ((\<lambda>s. shd s = False) \<circ> (sdrop n)) (events bern)) = (IN (\<lambda>s.  shd s = False) (events bern))"
    by (simp only: EVENTS_BERN_SDROP)
  ultimately have "IN (GSPEC (\<lambda>s. (s, (shd (sdrop n s) = False)))) (events bern)"
    by (simp add: GSPEC_DEF_ALT comp_def)
  hence "P (GSPEC (\<lambda>s. (s, (shd (sdrop n s) = False)))) = 1 - P (COMPL (GSPEC (\<lambda>s. (s, (shd (sdrop n s) = False)))))"
    using PROB_SPACE_BERN
    by (simp add: MEASURE_COMPL prob_space_def prob_def events_def)
  also have "(COMPL (GSPEC (\<lambda>s. (s, (shd (sdrop n s) = False))))) = (GSPEC (\<lambda>s. (s, \<not> (shd (sdrop n s) =  False))))"
    by (simp only: GCOMPL)
  also have "P (GSPEC (\<lambda>s. (s, \<not> (shd (sdrop n s) =  False)))) = 1 / 2"
    using PROB_BERN_SINGLE by simp
  finally have "P (GSPEC (\<lambda>s. (s, shd (sdrop n s) = False))) = 1/2"
    by simp
  with False show ?thesis
    by (simp add: GSPEC_lambda)
qed



lemma stake_prob: "\<And>r m. Suc r \<le> n \<Longrightarrow> P (\<lambda>s. stake n (sdrop m s)!r = b) = 1/2"
proof (induct n) 
  case 0 thus ?case by simp
next
  case (Suc n r m)
  thus ?case 
  proof (cases "r=0")
    case True
    hence "\<And>s. stake (Suc n) (sdrop m s)!r = shd (sdrop m s)" 
      by simp
    hence "\<And>s. (stake (Suc n) (sdrop m s)!r = b) = (shd (sdrop m s)=b)"
      by fast
    hence "(\<lambda>s. stake (Suc n) (sdrop m s)!r = b) = (\<lambda>s. shd (sdrop m s) = b)" 
      by (rule ext)
    thus ?thesis using shd_sdrop_prob
      by simp
  next
    case False
    then obtain u where u:"r = Suc u" 
      by (auto simp add: gr0_conv_Suc)
    hence "\<And>s. stake (Suc n) (sdrop m s)!r = stake n (stl (sdrop m s))!u" 
      by simp
    also
    { fix s
      have "stake n (stl (sdrop m s))!u = stake n (sdrop (Suc m) s)!u" 
	by (simp only: sdrop_stl[THEN sym] sdrop_def comp_def)
    }
    finally
    have "{s. stake (Suc n) (sdrop m s) ! r = b} = 
      {s. stake n (sdrop (Suc m) s) ! u = b}"
      by fast
    hence "\<And>s. (stake (Suc n) (sdrop m s) ! r = b) = 
      (stake n (sdrop (Suc m) s) ! u = b)"
      by fast
    hence "(\<lambda>s. stake (Suc n) (sdrop m s) ! r = b) = 
      (\<lambda>s. stake n (sdrop (Suc m) s) ! u = b)"
      by simp
    also
    from u Suc have "Suc u \<le> n" by simp
    with Suc have "P (\<lambda>s. stake n (sdrop (Suc m) s) ! u = b) = 1 / 2" 
      by fast
    ultimately show ?thesis 
      by simp
  qed
qed
   
text{*This generalizes the simple probability fact from the last
  section and almost immediately results in the probability for a single
  bit in the @{text lsdest} list.*}
  
theorem lsdest_prob: 
  assumes le: "Suc r \<le> n"
  shows "P (\<lambda>s. (fst (lsdest n s))!r = b) = 1/2" 
(*<*)proof -
  from le have "P (\<lambda>s. stake n (sdrop 0 s)!r = b) = 1/2"
    by (rule stake_prob[of r n 0 b])
  thus ?thesis 
    by (simp add: lsdesttakedrop)
qed(*>*)

text{*Another generalization, this time of the recursive @{text
  lsdest} definition, enables the decisive theorem.*}

lemma lsdest_split: 
  assumes le: "r \<le> n"
  shows "lsdest n = BIND (lsdest r) (\<lambda>(l::'a list). 
  BIND (lsdest (n-r)) (\<lambda>k. UNIT (l@k)))" (*<*)using le
proof (induct r)
  case 0 thus ?case 
    by (simp add: BIND_def UNIT_def split_def del: split_comp_eq)
next
  case (Suc u)
  hence "lsdest n = BIND (lsdest u) (\<lambda>(l::'a list). BIND (lsdest (n - u)) (\<lambda>k. UNIT (l @ k)))" 
    by simp
  also 
  { from prems have "n - u = Suc (n - Suc u)" by arith }
  hence "\<dots> = BIND (lsdest u) (\<lambda>(l::'a list). BIND (lsdest (Suc (n - Suc u))) (\<lambda>k. UNIT (l @ k)))"
    by simp
  also 
  have "\<dots> = BIND (lsdest u) (\<lambda>(l::'a list). 
    BIND (BIND sdest (\<lambda>x. BIND (lsdest (n - Suc u)) (\<lambda>r. UNIT (x#r)))) (\<lambda>k. UNIT (l @ k)))"
    by simp
  also
  
  { fix s
    have "\<dots> s = BIND (BIND sdest (\<lambda>x. BIND (lsdest u) (\<lambda>r. UNIT (x#r)))) 
      (\<lambda>(l::'a list). BIND (lsdest (n - Suc u)) (\<lambda>k. UNIT (l @ k))) s"
      by (simp add: BIND_def UNIT_def split_def lsdesttakedrop sdest_def sdrop_stl stake_shd del: split_comp_eq) }
  
  hence "\<dots> = BIND (lsdest (Suc u)) (\<lambda>(l::'a list). BIND (lsdest (n - Suc u)) (\<lambda>k. UNIT (l @ k)))"
    by (simp add: ext)
  
  finally show ?case .
qed(*>*)
    
text{*The upcoming powerful result closes the theory. Notice that the proposition
  had to be strengthened to arrive at a viable induction
  hypothesis. That is why a simplified version is attached.*}

lemma lsdest_prob2:
  shows "\<And>S n m. \<lbrakk>finite S; card S = k; \<forall>r\<in>S. Suc (r-m) \<le> n \<and> m \<le> r\<rbrakk>
  \<Longrightarrow> P (\<lambda>s. \<forall>r\<in>S. (fst (lsdest n s))!(r-m) = b r) = (1/2)^k"
proof (induct k)
  case (0 S)
  hence "S={}" 
    by simp
  thus ?case using PROB_BERN_UNIV 
    by (simp add: UNIV_def)
  
next
  case (Suc k S n m)
  then obtain r' where "r' \<in> S" 
    by force
  hence "(LEAST r. r \<in> S) \<in> S"
    by (rule LeastI)
  also 
  have "\<And>y. y\<in>S \<Longrightarrow> (LEAST r. r \<in> S) \<le> y" 
    by (rule Least_le)
  moreover 
  def r \<equiv> "(LEAST r. r \<in> S)"
  ultimately have rS:"r \<in> S" and "\<forall>y\<in>S. r \<le> y"
    by simp_all
  hence rl:"\<forall>y\<in>S-{r}. r < y" 
    by auto
  
  from rS have "S = insert r (S-{r})" 
    by auto 
  with rl have Sr2:"S = insert r {r2. r2 \<in> S \<and> r < r2}" 
    by auto
  hence "(\<lambda>s. \<forall>r\<in>S.  (fst (lsdest n s))!(r-m) = b r) = 
    (\<lambda>s. \<forall>r\<in> insert r {r2. r2 \<in> S \<and> r < r2}.
    (fst (lsdest n s))!(r-m) = b r)" 
    by simp
  also have "\<dots> = (\<lambda>s. (fst (lsdest n s))!(r-m) = b r \<and> 
    (\<forall>r2 \<in> {r2. r2 \<in> S \<and> r < r2}. (fst (lsdest n s))!(r2-m) = b r2))"
    by simp
  finally
  have "P (\<lambda>s. \<forall>r\<in>S. fst (lsdest n s) ! (r-m) = b r) =
    P (\<lambda>s. fst (lsdest n s) ! (r-m) = b r \<and>
    (\<forall>r2\<in>{r2. r2 \<in> S \<and> r < r2}. fst (lsdest n s) ! (r2-m) = b r2))" 
    by simp  

  also
  from prems rS have le: "Suc (r-m) \<le> n" 
    by simp
  { fix s
    from le have "fst (lsdest n s)!(r-m) = 
      fst (BIND (lsdest (Suc (r-m))) 
      (\<lambda>l. BIND (lsdest (n-(Suc (r-m)))) (\<lambda>k. UNIT (l@k))) s)!(r-m)"
      by (simp add: lsdest_split)
    hence "fst (lsdest n s)!(r-m) =  fst (lsdest (Suc (r-m)) s)!(r-m)"
      by (simp add: nth_append UNIT_def BIND_def split_def lsdest_length del: lsdest.simps split_comp_eq)
  } 
  hence "\<And>s. (fst (lsdest n s)!(r-m) = b r) = 
     (fst (lsdest (Suc (r-m)) s)!(r-m) = b r)"
    by fast
  moreover
  from prems rS have mr:"m \<le> r" 
    by simp
  
  { fix s r2 assume "r < r2" 
    with mr have l2: "\<not> (r2-m) < Suc (r-m)" by arith
    from le have "fst (lsdest n s) ! (r2-m) =
      fst (BIND (lsdest (Suc (r-m))) 
      (\<lambda>l. BIND (lsdest (n-(Suc (r-m)))) (\<lambda>k. UNIT (l@k))) s)!(r2-m)"
      by (simp add: lsdest_split)
    with l2 have "fst (lsdest n s) ! (r2-m) = 
      fst (lsdest (n - Suc (r-m)) 
      (snd (lsdest (Suc (r-m)) s))) ! ((r2 - m)  - Suc (r-m))"
      by (simp add: nth_append UNIT_def BIND_def split_def lsdest_length del: lsdest.simps split_comp_eq)
  }
  hence "\<And>s. (\<forall>r2\<in>{r2. r2 \<in> S \<and> r < r2}. fst (lsdest n s) ! (r2-m) = b r2) = 
    (\<forall>r2\<in>{r2. r2 \<in> S \<and> r < r2}. fst (lsdest (n - Suc (r-m)) 
    (snd (lsdest (Suc (r-m)) s))) ! ((r2-m) - Suc (r-m)) = b r2)"
    by fast
  ultimately have "P (\<lambda>s. \<forall>r\<in>S. fst (lsdest n s) ! (r-m) = b r) =
    P (\<lambda>s. fst (lsdest (Suc (r-m)) s)!(r-m) = b r \<and> 
    (\<forall>r2\<in>{r2. r2 \<in> S \<and> r < r2}. fst (lsdest (n - Suc (r-m)) 
    (snd (lsdest (Suc (r-m)) s))) ! ((r2-m) - Suc (r-m)) = b r2))"
    by simp  
  also 
  have "\<dots> = P (INTER ((\<lambda>l. l!(r-m) = b r) \<circ> (fst \<circ> lsdest (Suc (r-m)))) 
    ((\<lambda>s. \<forall>r2\<in>{r2. r2 \<in> S \<and> r < r2}. 
    fst (lsdest (n - Suc (r-m)) s) ! ((r2-m) - Suc (r-m)) = b r2) \<circ> (snd \<circ> lsdest (Suc (r-m)))))"
    by (simp add: INTER_def GSPEC_DEF_ALT IN_def)
  also 
  {  from prems Sr2 have "finite (insert r {r2. r2 \<in> S \<and> r < r2})"
      by simp
    hence fin2:"finite {r2. r2 \<in> S \<and> r < r2}" 
      by (simp only: finite_insert)
    hence "IN (\<lambda>s. \<forall>r2\<in>{r2. r2 \<in> S \<and> r < r2}. 
      fst (lsdest (n - Suc (r-m)) s) ! ((r2-m) - Suc (r-m)) = b r2) (events bern)"
    proof (induct "{r2. r2 \<in> S \<and> r < r2}")
      case empty 
      thus ?case using  EVENTS_BERN_UNIV
	by (simp add: UNIV_def)
    next
      case (insert F x)
      hence "(\<lambda>s. \<forall>r2\<in>insert x F. fst (lsdest (n - Suc (r - m)) s) ! (r2 - m - Suc (r - m)) = b r2)
	= (\<lambda>s. (\<forall>r2\<in>F. fst (lsdest (n - Suc (r - m)) s) ! (r2 - m - Suc (r - m)) = b r2) 
	\<and> fst (lsdest (n - Suc (r - m)) s) ! (x - m - Suc (r - m)) = b x)"
	by (simp add: conj_commute) 
      also have "\<dots> = INTER (\<lambda>s. (\<forall>r2\<in>F. fst (lsdest (n - Suc (r - m)) s) ! (r2 - m - Suc (r - m)) = b r2))
	(\<lambda>s. fst (lsdest (n - Suc (r - m)) s) ! (x - m - Suc (r - m)) = b x)"
	by (simp add: INTER_def GSPEC_DEF_ALT IN_def)
      also 
      { fix r2 
	from lsdest_indep[of "(n - Suc (r-m))"]
	have "IN ((\<lambda>l. l! ((r2-m) - Suc (r-m)) = b r2) \<circ> (fst \<circ> lsdest (n - Suc (r-m)))) (events bern)"
	  by (simp add: INDEP_FN_FST_EVENTS )
	hence "IN (\<lambda>s.  fst (lsdest (n - Suc (r-m)) s) ! 
	  ((r2-m) - Suc (r-m)) = b r2) (events bern)"
	  by (simp add: comp_def) 
      }
      ultimately show ?case using insert
	by (simp add: PROB_SPACE_BERN EVENTS_INTER) 
    qed
  }
  with lsdest_indep [of "Suc (r-m)"] 
  have "\<dots> = P ((\<lambda>l. l!(r-m) = b r) \<circ> (fst \<circ> lsdest (Suc (r-m)))) *
    P ((\<lambda>s. \<forall>r2\<in>{r2. r2 \<in> S \<and> r < r2}. 
    fst (lsdest (n - Suc (r-m)) s) ! ((r2-m) - Suc (r-m)) = b r2))"
    by (simp add: INDEP_FN_PROB)
  also
  { from le_refl[of "Suc (r-m)"] have 
    "P (\<lambda>s. fst (lsdest (Suc (r-m)) s)!(r-m) = b r) = 1/2" 
      by (rule lsdest_prob) }
  hence  "\<dots> = 1/2 * 
    P ((\<lambda>s. \<forall>r2\<in>{r2. r2 \<in> S \<and> r < r2}. 
    fst (lsdest (n - Suc (r-m)) s) ! ((r2-m) - Suc (r-m)) = b r2))"
    by (simp add: comp_def)
  
  also 
  { from mr have "\<And>r2. r2\<in>{r2. r2 \<in> S \<and> r < r2} \<Longrightarrow>  
    (r2-m) - Suc (r-m) = r2 - Suc r" 
      by arith
    hence "((\<lambda>s. \<forall>r2\<in>{r2. r2 \<in> S \<and> r < r2}. 
    fst (lsdest (n - Suc (r-m)) s) ! ((r2-m) - Suc (r-m)) = b r2))
      = ((\<lambda>s. \<forall>r2\<in>{r2. r2 \<in> S \<and> r < r2}. 
      fst (lsdest (n - Suc (r-m)) s) ! (r2 - Suc r) = b r2))"
      by simp
  }
  hence "\<dots> = 1/2 * P ((\<lambda>s. \<forall>r2\<in>{r2. r2 \<in> S \<and> r < r2}. 
      fst (lsdest (n - Suc (r-m)) s) ! (r2 - Suc r) = b r2))" 
    by simp

  also
  from prems Sr2 have "finite (insert r {r2. r2 \<in> S \<and> r < r2})"
    by simp
  hence fin2:"finite {r2. r2 \<in> S \<and> r < r2}" 
    by (simp only: finite_insert)
  with  prems Sr2 have ?this and "card (insert r {r2. r2 \<in> S \<and> r < r2}) = Suc k"
    by simp_all
  hence c2:"card {r2. r2 \<in> S \<and> r < r2} = k" 
    by (simp add: card_insert)

  { fix d assume "d \<in> {r2. r2 \<in> S \<and> r < r2}"
    hence "d \<in> S" and rd:"r < d" 
      by simp_all
    with prems have i:"Suc (d - m) \<le> n" and ii:"m \<le> d" 
      by simp_all
    from i rd mr ii have "Suc (d- Suc r) \<le> (n-Suc (r-m)) \<and> Suc r \<le> d" 
      by arith
  }
  hence "\<forall>d\<in>{r2. r2 \<in> S \<and> r < r2}. 
    Suc (d-(Suc r)) \<le> (n-Suc (r-m)) \<and> Suc r \<le> d"
    by simp

  with fin2 c2 prems have nn:"0<k \<Longrightarrow> P ((\<lambda>s. \<forall>r2\<in>{r2. r2 \<in> S \<and> r < r2}. 
    fst (lsdest (n - Suc (r-m)) s) ! (r2 - Suc r) = b r2)) = (1/2)^k"
    by (simp only:)
    
  have "P ((\<lambda>s. \<forall>r2\<in>{r2. r2 \<in> S \<and> r < r2}. 
    fst (lsdest (n - Suc (r-m)) s) ! (r2 - Suc r) = b r2)) = (1/2)^k"
  proof (cases k)
    case 0 
    with fin2 c2 have "{r2. r2 \<in> S \<and> r < r2} = {}"
      by simp
    hence "(\<lambda>s. \<forall>r2\<in>{r2. r2 \<in> S \<and> r < r2}. 
    fst (lsdest (n - Suc (r-m)) s) ! (r2 - Suc r) = b r2) = UNIV"
      by (simp only: UNIV_def) simp
    hence "P (\<lambda>s. \<forall>r2\<in>{r2. r2 \<in> S \<and> r < r2}. 
    fst (lsdest (n - Suc (r-m)) s) ! (r2 - Suc r) = b r2) = 1"
      by (simp add: PROB_BERN_UNIV)
    with 0 show ?thesis by simp
  next
    case (Suc g)
    hence "0<k" by simp
    with nn show ?thesis by simp
  qed

  finally show ?case by simp
qed


lemma lsdest_probs:
  shows "\<lbrakk>finite R; card R = k; \<forall>r\<in>R. Suc r \<le> n\<rbrakk>
  \<Longrightarrow> P (\<lambda>s. \<forall>r\<in>R. (fst (lsdest n s))!r = b r) = (1/2)^k"
(*<*)proof -
  assume fin: "finite R" and k: "card R = k"
    and S: "\<forall>r\<in>R. Suc r \<le> n"
  from S have "\<forall>r\<in>R. Suc (r - 0) \<le> n \<and> 0 \<le> r" 
    by simp
  with fin k 
  have "P (\<lambda>s. \<forall>r\<in>R. (fst (lsdest n s))!(r - 0) = b r) = (1/2)^k"
    by (rule lsdest_prob2)
  thus "P (\<lambda>s. \<forall>r\<in>R. (fst (lsdest n s))!r = b r) = (1/2)^k" 
    by simp
qed(*>*)
    
 
end
