header {*An Integral Example: The First Moment Method for kSAT*}

theory kSAT = ImportPredSet+Integral+Lsdest:

text{*As you can see from the @{text theory} line, we finally get to
  use the integral along with the new primitive just
  developed. Surprisingly, the end result will be a combinatorial
  statement about the existence of an assignment for a $k$-SAT
  instance. Notwithstanding, probability and expectation properties of
  some randomized programs are going to lead us there. The
  technique employed --- known as the first moment method --- is a
  standard one in the field of randomized algorithms; it may be found 
  in the authoritative textbook on the subject by Motwani and
  Raghavan \cite{MotRag97}. 

  Let the problem be defined first.*}
  (*kSAT*)

(*Just to reinstate the Set INTER definition*)
constdefs 
  INTER         :: "'a set => ('a => 'b set) => 'b set"  -- "general intersection"
  "INTER E F      == {y. ALL x:E. y: F(x)}" 

declare INTER_def [simp]

lemma INT_cong [cong]:
    "A = F ==> (!!x. x:F ==> E x = D x) ==> (INT x:A. E x) = (INT x:F. D x)"
  by (simp)

consts 
  absdistinct:: "int list \<Rightarrow> bool"
  clauseeval:: "int list \<Rightarrow> bool list \<Rightarrow>  bool"
  CNFeval:: "(int list) list \<Rightarrow> bool list \<Rightarrow> bool"

primrec
  "absdistinct [] = True"
  "absdistinct (x#xs) = (x \<notin> List.set xs \<and> -x \<notin> List.set xs \<and> absdistinct xs)" 

constdefs
  nvarkclauses:: "nat \<Rightarrow> nat \<Rightarrow> int list set" ("_ var _ clauses")
  "n var k clauses \<equiv> {ls. length ls = k \<and> 
  (\<forall>x \<in> List.set ls. x\<noteq>0 \<and> \<bar>x\<bar> \<le> int n) \<and> absdistinct ls}"
  
  nvarksat:: "nat \<Rightarrow> nat \<Rightarrow> ((int list) list) set" ("_ var _ SAT")
  "n var k SAT \<equiv> {A. \<forall>ls \<in> List.set A. ls \<in> n var k clauses}"

lemma assumes "a \<in> List.set D" and "b \<in> List.set D"
  and "absdistinct D" and "-a\<noteq>a"
  shows clause_dist: "-a \<noteq> b" using prems
  by (induct D) auto
 
lemma clauses_card:
  assumes  D: "absdistinct D"
  shows "card (List.set D) = length D" using D
  by (induct D) auto

text{*Thus formulas are modeled as lists of clauses, which in turn are
  represented by lists of integers. The absolute value of a number
  stands for the variable name, the sign signifying negation of the
  literal. For an illustrative instance,\nolinebreak $-4$ means the 4th variable
  inverted, and 0 is not allowed. A variable may not appear twice in
  any clause, as ensured by the @{text absdistinct} predicate. Looking
  at the following example might clarify the notation. *}

theorem "[[1,2,3],[-4,-2,1]] \<in> 4 var 3 SAT" 
  by (simp add: nvarksat_def nvarkclauses_def) 

text{*Formulas can be evaluated at an assignment, that
  is a list of booleans, by primitive recursive functions.*}

primrec
  "clauseeval [] l  = False"
  "clauseeval (x#xs) l  = (if (0<x) then (l!nat (x+-1) \<or> clauseeval xs l)
                           else if (x<0) then (\<not>(l!nat (-1+-x)) \<or> clauseeval xs l)
                                else True)"
 primrec
  "CNFeval [] l = True"
  "CNFeval (x#xs) l = (clauseeval x l \<and> CNFeval xs l)"

lemma CNF_clause: "CNFeval F l = (\<forall>C\<in>(List.set F). clauseeval C l)"
  by (induct F) auto

text{*Now we may randomize these functions, obtaining just the simple
  programs described in \ref{sec:lsdest}. In addition, an indicator
  variable is defined that takes the value $1$ for exactly those
  elementary events --- or rather bit sequences --- where the
  argument clause is not satisfied.*}

constdefs
  randCNFeval:: "(int list) list \<Rightarrow> nat \<Rightarrow> bool seq \<Rightarrow> (bool * (bool seq))"
  "randCNFeval F n s \<equiv> (CNFeval F (stake n s), sdrop n s)"
  (* These are our first randomized algorithms! *)
  randclauseeval:: "int list \<Rightarrow> nat \<Rightarrow> bool seq \<Rightarrow> (bool * (bool seq))"
  "randclauseeval D n \<equiv> BIND (lsdest n) (\<lambda>l. UNIT (clauseeval D l))"

  indicator:: "int list \<Rightarrow> nat \<Rightarrow> bool seq \<Rightarrow> real"
  "indicator D n \<equiv> \<chi>{s. \<not> fst (randclauseeval D n s)}"

lemma randCNFeval_BIND_UNIT: "randCNFeval F n s = 
  BIND (lsdest n) (\<lambda>l. UNIT (CNFeval F l)) s" 
  by (simp add: randCNFeval_def BIND_def UNIT_def lsdesttakedrop)      

lemma rand_CNF_clause: "fst (randCNFeval F n s) = 
  (\<forall>C\<in>List.set F. fst (randclauseeval C n s))"
  by (simp add: 
    CNF_clause randCNFeval_BIND_UNIT randclauseeval_def BIND_def UNIT_def split_def del: split_comp_eq)

text {*We just saw that both @{text randclauseeval} and @{text randCNFeval} can be built from @{text UNIT}, @{text BIND} and @{text sdest} alone. Hence they are strongly independent
  functions. In particular, indicator is a characteristic function
  for an event. *}

lemma rce_indep: "IN (randclauseeval D n) indep_fn"
  by (simp add: lsdest_indep INDEP_FN_BIND INDEP_FN_UNIT randclauseeval_def)

lemma rce_events: "IN (\<lambda>s. \<not> fst (randclauseeval D n s)) (probability.events prob.bern)"
proof -
  from rce_indep have "IN (Not \<circ> (fst \<circ> randclauseeval D n)) (probability.events prob.bern)"
    by (simp only: INDEP_FN_FST_EVENTS)
  thus ?thesis  
    by (simp add: comp_def)
qed

text{*The next step is to compute the measure of this event, the
  probability that a given clause is not satisfied. In spite of the
  preparatory work on @{text lsdest}, the greatest difficulty lies in
  here. Though a rough idea should have emerged until now, it is
  technically demanding to arrive at a setup where
  @{text lsdest_probs} may be applied instantly. No general insight
  is gained from the proofs, so they are left out.*}

lemma clauseeval_inj:
  assumes cont: "(\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) p 
  = (\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) r"
  and r0: "r \<noteq> 0" and p0: "p \<noteq> 0"
  shows "p = r \<or> -p = r"
(*<*)proof (cases "0< p")
  case True
  assume h0: "0<p"
  show ?thesis
  proof (cases "0<r")
    case True
    with h0 cont
    have "int (nat (p + -1)) = int (nat (r + -1))"
      by simp
    with h0 True show ?thesis
      by simp
  next
    case False
    with h0 cont
    have z: "int (nat (p + -1)) = int (nat (-1 - r))"
      by simp
    from False r0 have "0\<le> -1 - r" 
      by arith
    with h0 z have "(p + -1) = (-1 - r)"
      by simp
    thus ?thesis by simp
  qed
next
  case False
  with p0 have h0: "p < 0" 
    by arith
  show ?thesis
  proof (cases "0<r")
    case True
    with h0 cont
    have "int (nat (-1 - p)) = int (nat (r + -1))"
      by simp
    with h0 True show ?thesis
      by simp
  next
    case False
    with h0 cont
    have z: "int (nat (- 1 - p )) = int (nat (-1 - r))"
      by simp
    from False r0 have "0\<le> -1 - r" 
      by arith
    with h0 z have "(-1 - p) = (-1 - r)"
      by simp
    thus ?thesis by simp
  qed
qed(*>*)


lemma not_clauseeval: 
  "\<And>D. D \<in> n var k clauses \<Longrightarrow> 
  \<exists>b. \<forall>l. (\<not> clauseeval D l) = 
  (l \<in> (\<Inter>r\<in>((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`List.set D). 
  {l. l!r = b r}))" 
(*<*)proof (induct k)
  case 0 thus ?case 
    by (simp add: nvarkclauses_def)
next
  case (Suc k D)
  
  hence "D \<noteq> []"
    by (auto simp add: nvarkclauses_def)
  hence ht: "D = hd D#tl D" 
    by simp
  
  have "hd D \<in> List.set (hd D#tl D)" 
    by simp
  with Suc ht have  n0: "hd D \<noteq> 0" 
    by (auto simp add: nvarkclauses_def)
  
  from Suc ht have aht:"absdistinct (hd D#tl D)" 
    by (simp add: nvarkclauses_def)
  hence "absdistinct (tl D)" 
    by simp
  
  also
  have "List.set (tl D) \<subseteq> List.set (hd D#tl D)"  
    by (rule set_subset_Cons)
  with ht have "List.set (tl D) \<subseteq> List.set D" 
    by simp

  moreover note Suc 
  
  ultimately
  have tc: "tl D \<in> n var k clauses" 
    by (auto simp add: nvarkclauses_def)
  with Suc have "\<exists>b. \<forall>l. (\<not> clauseeval (tl D) l) = 
    (l \<in> (\<Inter>r\<in> ((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`List.set (tl D)). {l.  l!r = b r}))"
    by simp
  then obtain b where b: "\<forall>l. (\<not> clauseeval (tl D) l) = 
    (l \<in> (\<Inter>r\<in> (\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`List.set (tl D). {l.  l!r = b r}))"
    by auto
  
  have "\<exists>c. \<forall>l. (\<not> clauseeval D l) = 
    ((l!(\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D) = c) \<and> \<not> clauseeval (tl D) l)"
  proof (cases "0 < hd D")
    case True
    hence "\<forall>l. (\<not> clauseeval (hd D#tl D) l) = 
      ((l!(\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D) = False) 
      \<and> \<not> clauseeval (tl D) l)"
      by simp
    with ht show ?thesis by auto
  next
    case False
    with n0 have "hd D < 0" 
      by arith
    hence "\<forall>l. (\<not> clauseeval (hd D#tl D) l) =
      ((l!(\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D) = True) 
      \<and> \<not> clauseeval (tl D) l)"
      by simp
    with ht show ?thesis by auto
  qed
  then obtain c where c: "\<forall>l. (\<not> clauseeval D l) =
    (l ! (\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D) = c \<and> \<not> clauseeval (tl D) l)" 
    by auto

  { fix l
    from b c have "(\<not> clauseeval D l) = 
      ((l ! (\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D) = c) \<and> 
      (l \<in> (\<Inter>r\<in>(\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`List.set (tl D). 
      {l.  l!r = b r})))" 
      by simp
    
    also
    { from aht have ne:"hd D \<notin> List.set (tl D) \<and> - hd D \<notin> List.set (tl D)" 
	by simp
      { fix r assume rD: "r \<in> List.set (tl D)"
	and cont: "(\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D) 
	= (\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) r"
	
	from rD tc have r0:"r \<noteq> 0" 
	  by (simp add: nvarkclauses_def)
	with cont n0 have "(hd D) = r \<or> -(hd D) = r"
	  by (simp add: clauseeval_inj)
	with rD have
	  "hd D \<in> List.set (tl D) \<or>  - hd D \<in> List.set (tl D)"
	  by auto
	with ne have False by simp
      }
      hence "\<And>r. r \<in> (\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`List.set (tl D)
	\<Longrightarrow> r \<noteq> (\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D)"
	by force

      hence "\<And>r. r \<in> (\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`List.set (tl D) \<Longrightarrow>  
	b r = (b((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D):= c)) r" 
	by simp
      hence "\<And>r. r \<in> (\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`List.set (tl D) \<Longrightarrow>
	{l.  l!r = b r} = 
	{l.  l!r = (b((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D):= c)) r}"
	by  simp
      
      with refl have "(\<Inter>r\<in>(\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`List.set (tl D).
	{l.  l!r = b r}) =
	(\<Inter>r\<in>(\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`List.set (tl D).
	{l.  l!r = (b((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D):= c)) r})"
	by (rule INT_cong)     
      
    }
    hence "\<dots> = ((l ! (\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D) = 
      (b((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D):= c)) 
      ((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D))) \<and> 
      (l \<in> (\<Inter>r\<in>(\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`List.set (tl D). 
      {l.  l!r = (b((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D):= c)) r})))" 
      by simp
    
    also 
    have "\<dots> = (l \<in> {l. l ! (\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D) = 
      (b((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D):= c)) 
      ((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D))} \<and> 
      (l \<in> (\<Inter>r\<in>(\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`List.set (tl D). 
      {l.  l!r = (b((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D):= c)) r})))" 
      by simp
    
    also
    have "\<dots> = (l \<in> (\<Inter>r\<in> 
      insert ((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D)) 
      ((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`(List.set (tl D))). 
      {l.  l!r = (b((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D):= c)) r}))"
      by simp
 
    also   
    { have "insert ((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D)) 
      ((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`(List.set (tl D))) =
      (\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`
      insert (hd D) (List.set (tl D))"
	by (rule image_insert[THEN sym]) 
    }
    hence "\<dots> = (l \<in> (\<Inter>r\<in> (\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`
      insert (hd D) (List.set (tl D)). {l.  l!r =
      (b((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D):= c)) r}))"
      by simp
    
    also
    from ht have 
      "\<dots> =  (l \<in> (\<Inter>r\<in> (\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`(List.set D). 
      {l.  l!r =
      (b((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (hd D):= c)) r}))"
      by (simp add: set.simps[THEN sym])
  
    finally have "(\<not> clauseeval D l) = \<dots>" .
  }
    
  thus ?case by blast
qed(*>*)

theorem assumes D: "D \<in> n var k clauses" 
  shows rce_prob: "Lsdest.P (\<lambda>s. \<not> fst (randclauseeval D n s)) = (1/2)^k"
(*<*)proof -
  have "(\<lambda>s. \<not> fst (randclauseeval D n s)) =
    (\<lambda>s. \<not> clauseeval D (fst(lsdest n s)))" 
    by (simp add: randclauseeval_def BIND_def UNIT_def split_def del: split_comp_eq)
  
  also 
  from D have "\<exists>b. \<forall>l.
    (\<not> clauseeval D l) = 
    (l \<in> (\<Inter>r\<in>((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`List.set D). {l. l!r = b r}))"
    by (rule not_clauseeval)
  then obtain b where 
    "\<And>l. (\<not> clauseeval D l) =
    (l \<in> (\<Inter>r\<in>((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`List.set D). {l. l!r = b r}))"
    by auto 
  hence "(\<lambda>s. \<not> clauseeval D (fst(lsdest n s))) = 
    (\<lambda>s. (fst(lsdest n s)) \<in> 
    (\<Inter>r\<in>((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`List.set D). {l. l!(r-(0::nat)) = b r}))"
    by simp
  
  also have "\<dots> = (\<lambda>s. \<forall>r\<in>(\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`List.set D. 
    fst(lsdest n s)!(r - 0) = b r)" by auto

  finally have eq: "(\<lambda>s. \<not> fst (randclauseeval D n s)) =
    (\<lambda>s. \<forall>r\<in>(\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`List.set D. 
    fst(lsdest n s)!(r - 0) = b r)" .

  have fin: "finite (List.set D)" 
    by simp
  hence "finite ((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`List.set D)"
    by (rule finite_imageI)

  moreover
  { fix p r assume a: "p \<in> List.set D" and b: "r \<in> List.set D"
    and cont: "(\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) p 
    = (\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) r"
    from a b D have "r\<noteq>0" and p: "p\<noteq>0" 
      by (simp_all add: nvarkclauses_def)
    with cont have "p=r \<or> -p=r" 
      by (rule clauseeval_inj)
   
    also from p have p2:"-p\<noteq>p"
      by arith
    from D have "absdistinct D"
      by (simp add: nvarkclauses_def)
    from a b this p2 have "-p\<noteq>r" 
      by (rule clause_dist)

    ultimately have "p=r" by simp
  }
  hence io: "inj_on (\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) (List.set D)"
    by (simp add: inj_on_def)
  from D have cD: "card (List.set D) = k" 
    by (simp add: nvarkclauses_def clauses_card)
  
  from fin io have "card ((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`List.set D) =
    card (List.set D)" 
    by (rule card_image) 
  with cD have "card ((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`List.set D) = k"
    by simp

  moreover
  { fix x assume "x \<in> List.set D"
    with D have xintn:"\<bar>x\<bar> \<le> int n" and x0:"x \<noteq> 0"
      by (simp_all add: nvarkclauses_def)
    
    hence "Suc ((\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x)) x) \<le> n"
    proof (cases "0<x")
      case True
      with xintn have le: "x \<le> int n" 
	by arith
      from True have "0 < x \<or> 0 \<le> int n" 
	by simp
      hence "(nat x \<le> nat (int n)) = (x \<le> int n)"
	by (rule nat_le_eq_zle)
      with le have "nat x \<le> n"
	by simp
      also
      from True have "nat x = nat (1+(x+-1))"
	by simp
      with True have "nat x = Suc (nat (x+-1))"
	by (simp add: Suc_nat_eq_nat_zadd1)
      moreover note True
      ultimately 
      show ?thesis by simp
    next
      case False
      with x0 have less: "x<0" 
	by arith
      hence le: "0 \<le> (-1+-x)"
	by arith
      from less xintn
      have len: "-x \<le> int n" 
	by arith
      have "0 < -x \<or> 0 \<le> int n" 
	by simp
      hence "(nat (-x) \<le> nat (int n)) = (-x \<le> int n)"
	by (rule nat_le_eq_zle)
      with len have "nat (-x) \<le> n"
	by simp
      also 
      have "-x = 1+(-1+-x)"
	by arith
      with le have "nat (-x) = Suc (nat (-1+-x))"
	by (simp add: Suc_nat_eq_nat_zadd1)
      also
      note False
      ultimately
      show ?thesis by simp
    qed
  }
  hence "\<forall>x \<in> ((\<lambda>x. if 0 < x then nat (x + -1) else nat (-1 + - x)) ` List.set D). Suc x \<le> n"
    by force
  
  ultimately
  have "Lsdest.P (\<lambda>s. \<forall>r\<in>(\<lambda>x. if 0<x then nat (x+-1) else nat (-1+-x))`List.set D. 
    fst(lsdest n s)!r = b r) = (1/2)^k"
    by (rule lsdest_probs)
    
  with eq show ?thesis 
    by simp
qed(*>*)

thm rce_events
text{*We should take a moment to appreciate this first result. It
  embodies the gist of the probabilistic analysis for the @{text
  randclauseeval} randomized algorithm. What is more, it enables the
  primal application of integration in the following theorem.*}

thm integral_char

theorem assumes cl: "D \<in> n var k clauses" 
  shows ind_int: "\<integral> (indicator D n) \<partial> bern = (1/2)^k" 
  and "integrable (indicator D n) bern"
proof -
  have 1: "{s. \<not> fst (randclauseeval D n s)} \<in> events bern"
    using rce_events by (simp add: events_pred)
  also from 1 cl have "prob bern {s. \<not> fst (randclauseeval D n s)} = (1 / 2) ^ k"
    using rce_prob by (simp add: prob_pred)
  ultimately show "\<integral> (indicator D n) \<partial> bern = (1/2)^k"
    and "integrable (indicator D n) bern"
    using ps_bern
    by (auto simp add: indicator_def integral_char prob_space_def)
qed

text{*Here we encounter an expectation in the true sense for the first
  time in this thesis. Like any expectation it sums up easily.*}

lemma nat_lessThan_card:
  "card {..(k::nat)(} = k"
  by (induct k) (auto simp add: lessThan_Suc)
  
theorem sum_ind_int: 
  assumes sat: "F \<in> n var k SAT" 
  shows 
  "\<integral> (\<lambda>s. \<Sum>m\<in>{..(length F)(}. indicator (F!m) n s) \<partial>(Epsilon, P) 
  = real (length F)/2^k"
  and "integrable (\<lambda>s. \<Sum>m\<in>{..(length F)(}. indicator (F!m) n s)(Epsilon, P)" 
(*<*)proof -
  have f: "(\<lambda>s. \<Sum>m\<in>{..(length F)(}. indicator (F!m) n s) =
    (\<lambda>s. \<Sum>m\<in>{..(length F)(}. 1*\<chi>{s. \<not> fst (randclauseeval (F!m) n s)} s)"
    by (simp add: indicator_def)
  have fin: "finite {..(length F)(}" 
    by simp
  have ev: "\<forall>m \<in> {..(length F)(}. {s. \<not> fst (randclauseeval (F!m) n s)} 
    \<in> events bern" using rce_events
    by (simp add: events_pred)
  have nn: "nonnegative (\<lambda>i. 1::real)" 
    by (simp add: nonnegative_def)
  
  from ps_bern have ms_bern: "Measure.measure_space bern"
    by (simp add: prob_space_def)
  from this ev nn fin
  have "(\<Sum>m\<in>{..(length F)(}. 
    1*Measure.measure bern {s. \<not> fst (randclauseeval (F!m) n s)}) \<in> 
    sfis (\<lambda>s. \<Sum>m\<in>{..(length F)(}. 
    1*\<chi>{s. \<not> fst (randclauseeval (F!m) n s)} s) bern"
    by (rule sfis_intro)
  hence int: "(\<Sum>m\<in>{..(length F)(}. 
    Measure.measure bern {s. \<not> fst (randclauseeval (F!m) n s)}) \<in>
    nnfis (\<lambda>s. \<Sum>m\<in>{..(length F)(}. indicator (F!m) n s) bern"
    by (simp add: sfis_nnfis f)
  from int ms_bern 
  have "integrable (\<lambda>s. \<Sum>m\<in>{..(length F)(}. indicator (F!m) n s) bern"
    by (rule nnfis_integral)
  thus "integrable (\<lambda>s. \<Sum>m\<in>{..(length F)(}. indicator (F!m) n s) (Epsilon,P)"
    by (simp add: Epsilon_def P_def measure_def measurable_sets_def)

  from int ms_bern
  have "\<integral> (\<lambda>s. \<Sum>m\<in>{..(length F)(}. indicator (F!m) n s) \<partial> bern =
    (\<Sum>m\<in>{..(length F)(}. Measure.measure bern {s. \<not> fst (randclauseeval (F!m) n s)})"
    by (rule nnfis_integral)
  also
  { from sat have "\<And>m. m\<in>{..(length F)(} \<Longrightarrow> F!m \<in> n var k clauses"
      by (auto simp add: nvarksat_def)
    hence "IN (\<lambda>s. \<not> fst (randclauseeval (F!m) n s)) (probability.events prob.bern)"
      and "\<And>m. m\<in>{..(length F)(} \<Longrightarrow> Lsdest.P (\<lambda>s. \<not> fst (randclauseeval (F!m) n s)) = (1/2)^k"
      using rce_prob rce_events by auto 
    hence "\<And>m. m\<in>{..(length F)(} \<Longrightarrow> Measure.measure bern
      {s. \<not> fst (randclauseeval (F!m) n s)} = (1/2)^k"
      by (simp add: rce_events events_pred prob_pred)
  }
  with refl have "\<dots> = (\<Sum>m\<in>{..(length F)(}. (1/2)^k)"
    by (rule setsum_cong)
  
  also
  from fin have "\<dots> =  real (length F) * (1/2)^k" 
    by (simp add: nat_lessThan_card setsum_const_real)

  also 
  { have "(1/(2::real))^k = 1^k/2^k" 
      by (rule realpow_divide)
    hence "(1/(2::real))^k = 1/2^k" 
      by simp
  }
  hence "\<dots> = real (length F)/2^k" 
    by simp

  finally show "\<integral> (\<lambda>s. \<Sum>m\<in>{..(length F)(}. indicator (F!m) n s) \<partial>(Epsilon, P) 
    = real (length F)/2^k"
    by (simp add: Epsilon_def P_def Measure.measure_def Measure.measurable_sets_def)
qed(*>*)

text{*The result just obtained contributes all the information about
  probabilistic programs we will need: The expected number of
  unsatisfied clauses with our simplistic algorithm is the total
  number of clauses divided by $2^k$. It is only now that the first
  moment method comes into play. The point put forward by this
  proposition is that
  if the expected value of nonnegative random variable is less than
  $1$, then there must be an event witnessing this. The proof turns
  out to be rather elementary from the Markov inequation. *}

theorem first_moment_method: 
  assumes int1: "integrable f M" and nn: "nonnegative f" and int2: "\<integral> f \<partial> M < 1"
  shows "law M f {1..} < 1"
proof -
  from nn have eq: "f = (\<lambda>t. \<bar>f t\<bar>^1)" 
    by (simp_all add: nonnegative_def abs_eqI1)
  with int1 have "0<(1::real)" and "integrable (\<lambda>t. \<bar>f t\<bar>^1) M" 
    by simp_all
  with int1 have "law M f {1..} \<le> \<integral> (\<lambda>t. \<bar>f t\<bar>^1) \<partial> M / (1^1)"
    by (rule markov_ineq) 
  also from eq int2 have "\<dots> < 1"
    by simp
  finally show ?thesis .
qed
  
corollary fmm: 
  assumes int1: "integrable f M" and int2: "\<integral> f \<partial> M < 1"
  and ps: "prob_space M"
  shows "\<exists>s. f s < 1"
proof (cases "nonnegative f")
  case False txt{*\nopagebreak*}
  then obtain s where "\<not> 0\<le> f s" 
    by (auto simp add: nonnegative_def) 
  hence "f s < 1" by arith
  thus ?thesis by fast
next
  case True
  from int1 True int2 have "law M f {(1::real)..} < 1"
    by (rule first_moment_method)
  with int1 ps have "(f -` {(1::real)..}) \<noteq> {s. True}"
    by (auto simp add: integrable_rv prob_space_def distribution_def)
  then obtain s where "\<not> 1 \<le> f s" 
    by auto
  hence "f s < 1" by arith
  thus ?thesis 
    by fast
qed

text{*In the application we have in mind, a random bit-stream that
  makes the indicator variables sum to a value less than $1$
  corresponds to a satisfying assignment.*}

lemma assumes nk: "F \<in> n var k SAT" and
  sum: "(\<Sum>m\<in>{..(length F)(}. indicator (F!m) n env) < 1"
  shows satisfy: "CNFeval F (stake n env)"
proof -
  have fin: "finite {..(length F)(}" 
    by simp
  { 
    fix i assume "i \<in> {..(length F)(}"
      and "indicator (F!i) n env = 1"
    with sum fin have "(\<Sum>m\<in>{..(length F)(}-{i}. indicator (F!m) n env) < 0"
      by (simp add: setsum_diff_real)
    also have "\<forall>j\<in>{..(length F)(}-{i}. 0 \<le> indicator (F!j) n env"
      by (simp add: indicator_def characteristic_function_def)
    hence "0 \<le> (\<Sum>m\<in>{..(length F)(}-{i}. indicator (F!m) n env)"
      by (rule setsum_ge0_real)
    
    finally have False by simp
  }
  hence "\<forall>i \<in> {..(length F)(}. fst (randclauseeval (F ! i) n env)"
    by (auto simp add: indicator_def characteristic_function_def if_def)
  hence "\<forall>D \<in> List.set F. fst (randclauseeval D n env)"
    by (auto simp add: set_conv_nth)
  hence "fst (randCNFeval F n env)" 
    by (simp add: rand_CNF_clause)
  thus ?thesis 
    by (simp add: randCNFeval_def)
qed

text{*In the end we have shown that a satisfying assignment always
  exists if there are less than $2^k$ clauses in a $k$-CNF
  formula.*} 

theorem assumes nk:"F \<in> n var k SAT" and l:"real(length F) < 2^k"
  shows existence: "\<exists>l. CNFeval F l"
proof -
  from nk have int: 
    "integrable (\<lambda>s. \<Sum>m\<in>{..(length F)(}. indicator (F!m) n s) (Epsilon, P)"
    by (rule sum_ind_int)
  from nk have 
    "\<integral> (\<lambda>s. \<Sum>m\<in>{..(length F)(}. indicator (F!m) n s) \<partial>(Epsilon, P) 
    = real (length F)/2^k" 
    by (rule sum_ind_int)
  also 
  have "(0::real) < 2^k" 
    by (simp add: realpow_gt_zero)
  with l have "(real (length F)/2^k < 1)" 
    by (simp add: pos_real_divide_less_eq)
  finally have 
    "\<integral> (\<lambda>s. \<Sum>m\<in>{..(length F)(}. indicator (F!m) n s) \<partial>(Epsilon, P) < 1" .
  
  with int ps_bern obtain env where 
    "(\<Sum>m\<in>{..(length F)(}. indicator (F!m) n env) < 1" 
    by (force simp add: fmm P_def Epsilon_def measure_def measurable_sets_def) 
  with nk have "CNFeval F (stake n env)" 
    by (rule satisfy)
  thus ?thesis ..
qed

text{*The final result is a purely combinatorial fact, serving as an
  example for the application of probability and integration in an
  apparently unrelated field. There are stronger versions, requiring
  more intricate algorithms and tools\footnote{The Lovasz
  Local Lemma is a prominent example.}. Anyhow, this one should
  suffice to display the power of
  the basic integral properties.*}

end
